#include <string.h>
#include <stdio.h>

#include "TextView.h"

#include "gpmem.h"
#include "gpstdio.h"
#include "gpstdlib.h"

#include "Textout.h"
#include "Key.h"

#define TEXTVIEW_LEFT		7
#define TEXTVIEW_TOP		4
#define TEXTVIEW_ROW		14
#define TEXTVIEW_COL		38

#define TEXTVIEW_HEIGHT		16

#define TEXTVIEW_MAX_LINE	200000

static void (*pTextOutLXY)(int, int, char*, int, WORD);

typedef struct _TEXTLINE
{
	LPBYTE	pszText;
	int		nLength;
} TEXTLINE, *LPTEXTLINE;

struct
{
	int 	nViewInfo;
	BOOL	bWordWarp;

	int 	nTop;
} BookMark;

BOOL	gpMP3Back = FALSE;
BOOL	gpBookMark = FALSE;

static LPBYTE 		gpTextViewBuffer = NULL;
static DWORD		gpTextViewLength = 0;

static int         	gpTextLineCount = 0;

static char			gpTextViewFileName[256];

static LPTEXTLINE	gpTextViewLine;
static int			gpAllocCount;

static unsigned int	gpFreeMem;

static void DestroyTextView(void)
{
	if (gpTextViewLine)
	{
		gm_free(gpTextViewLine);
		gpTextViewLine = NULL;
	}

	if (gpTextViewBuffer)
	{
		gm_free(gpTextViewBuffer);
		gpTextViewBuffer = NULL;
	}
}

#define TAB_LENGTH 4
static int GetTextLine(LPBYTE lpszBuffer, int* lpLength)
{
	int nCount  = 0;
	int nLength = 0;

	static BOOL bNextLine = FALSE;

	LPBYTE pszBuffer = lpszBuffer;
	while (*pszBuffer != '\0')
	{
		BYTE code = *pszBuffer ++;

		if (code == 0x09)
		{
			int nSkip = TAB_LENGTH;
			if (nSkip > (TEXTVIEW_COL - nLength))
				nSkip = (TEXTVIEW_COL - nLength);

			nLength += (nSkip - 1);
		}
		else if (code == 0x0A)
		{
			if (BookMark.bWordWarp)
			{
				if (bNextLine)
				{
					bNextLine = FALSE;

					nLength ++;
					nCount ++;
					continue;
				}
			}

			*lpLength = nLength + 1;
			return (nCount + 1);
		}

		nCount  ++;
		nLength ++;

		///////////////////////////////////////////////////////////////////////
		if (nLength > (TEXTVIEW_COL - 1))
		{
			if (code & 0x80)
			{
				if (GetCodeType(lpszBuffer, nCount - 1) == CODE_FIRST)
				{
					nLength --;
					nCount  --;
				}
			}

			bNextLine = TRUE;
			break;
		}
	}

	*lpLength = nLength;

	return nCount;
}
static LPTEXTLINE AllocTextLine()
{
	LPTEXTLINE pLine;

	if (gpAllocCount > (TEXTVIEW_MAX_LINE - 1))
		return NULL;

	pLine = gpTextViewLine + gpAllocCount;
	gpAllocCount ++;

	return pLine;
}

static BOOL MakeTextLineBuffer()
{
	int 		nLength, nCount;
	DWORD		dwOffset = 0;

	LPBYTE 		pBuffer;
	LPTEXTLINE 	pLine;

	//
	pBuffer = gpTextViewBuffer;

	BookMark.nTop = 0;

	gpAllocCount = 0;
	gpTextLineCount = 0;

   	while (1)
   	{
		nCount = GetTextLine(pBuffer, &nLength);
		if (nLength)
		{
	      	pLine = AllocTextLine();
	      	if (!pLine)
	      		return FALSE;

			pLine->pszText = pBuffer;
			pLine->nLength = nLength;

			gpTextLineCount ++;
		}
		pBuffer += nCount;

		dwOffset += nCount;
		if (dwOffset > (gpTextViewLength - 1))
			break;
	}

	return TRUE;
}

static void MoveLine(int nRow)
{
	if (nRow < 0)
	{
		while (nRow ++ != 0 && BookMark.nTop > 0)
			BookMark.nTop --;
	}
	else
	{
		while (nRow -- != 0 && BookMark.nTop < (gpTextLineCount - TEXTVIEW_ROW))
			BookMark.nTop ++;
	}
}

static void LoadBookMark(void)
{
	char szFileName[256];
	char* ptr;

	F_HANDLE	fHandle;
	ERR_CODE 	result;

	DWORD  		dwDummy;

	//
	strcpy(szFileName, gpTextViewFileName);

	ptr = strchr(szFileName, '.');
	*ptr = '\0';

	strcat(ptr, ".SAV");

	//
	result = GpFileOpen(szFileName, OPEN_R, &fHandle);
	if (result != SM_OK)
	{
		BookMark.nViewInfo = 2;
		BookMark.bWordWarp = TRUE;

		BookMark.nTop = 0;

		gpBookMark = FALSE;

		return;
	}

	GpFileRead(fHandle, &BookMark, sizeof(BookMark), &dwDummy);
	GpFileClose(fHandle);

	gpBookMark = TRUE;
}
static void SaveBookMark(void)
{
	char szFileName[256];
	char* ptr;

	F_HANDLE	fHandle;
	ERR_CODE 	result;

	//
	strcpy(szFileName, gpTextViewFileName);

	ptr = strchr(szFileName, '.');
	*ptr = '\0';

	strcat(ptr, ".SAV");

	//
	result = GpFileCreate(szFileName, ALWAYS_CREATE, &fHandle);
	if (result != SM_OK)
	{
		gpBookMark = FALSE;
		return;
	}

	GpFileWrite(fHandle, &BookMark, sizeof(BookMark));
	GpFileClose(fHandle);
}

BOOL LoadTextView(const char* lpszFileName)
{
	int 	nRow;
	BOOL 	result;

	// Load text
	SetClockSpeed(CPU_CLOCK133);
	gpTextViewBuffer = GpLoadFromFile(lpszFileName, &gpTextViewLength);
	SetClockSpeed(CPU_CLOCK40);

	if (!gpTextViewBuffer)
		return FALSE;

	gpTextViewBuffer[gpTextViewLength] = 0;

	///////////////////////////////////////////////////////////////////////////
	strcpy(gpTextViewFileName, lpszFileName);

	//
	gpTextViewLine = (LPTEXTLINE) gm_malloc(TEXTVIEW_MAX_LINE * sizeof(TEXTLINE));
	if (!gpTextViewLine)
	{
		gm_free(gpTextViewBuffer);
		return FALSE;
	}

	gpFreeMem = gm_availablesize();

	// Auto detect hangul code
	if (IsKS(gpTextViewBuffer, gpTextViewLength))
		pTextOutLXY = GpKSTextOutLXY;
	else
		pTextOutLXY = GpKSSMTextOutLXY;

	//
	LoadBookMark();

	//
	nRow = BookMark.nTop;

	SetClockSpeed(CPU_CLOCK133);
	result = MakeTextLineBuffer();
	SetClockSpeed(CPU_CLOCK40);

	if (!result)
	{
		gm_free(gpTextViewLine);
		gm_free(gpTextViewBuffer);

		return FALSE;
	}

	//
	BookMark.nTop = nRow;

	return TRUE;
}

static void SetFont(void)
{
}

static void SetWordWarp(void)
{
	BookMark.bWordWarp ^= 1;

	if (!gpMP3Back)
		SetClockSpeed(CPU_CLOCK133);

	MakeTextLineBuffer();

	if (!gpMP3Back)
		SetClockSpeed(CPU_CLOCK40);
}

void ShowTextView(void)
{
	int  nRow = 0;
	char szText[80];

	LPTEXTLINE pLine = gpTextViewLine + BookMark.nTop;

	GpFillBlendMask75(0, 0, 320, 229, COLOR_BLUE);

	///////////////////////////////////////////////////////////////////////////
	if (gpMP3Back)
		UpdateMP3PlayListNextChunk();

	///////////////////////////////////////////////////////////////////////////
	GpFillBlendMask75(0, 229, 320, 11, COLOR_MAGENTA);

	// Line count
	GpTextOutXY(4, 229, gpTextViewFileName + 3, 0x0f);

	sprintf(szText, "%6d/%6d", BookMark.nTop, gpTextLineCount - 1);
	GpTextOutXY(240, 229, szText, 0x0f);

	if (gpMP3Back)
		GpTextOutXY(165, 229, GetMP3PlayList(), 0x0f);

	///////////////////////////////////////////////////////////////////////////
	if (gpMP3Back)
		UpdateMP3PlayListNextChunk();

	//
	while (nRow < TEXTVIEW_ROW)
	{
		if ((BookMark.nTop + nRow) > (gpTextLineCount - 1))
			break;

		pTextOutLXY(TEXTVIEW_LEFT, TEXTVIEW_TOP + nRow * TEXTVIEW_HEIGHT, (char *) pLine->pszText, pLine->nLength, COLOR_WHITE);
		pLine ++;

		nRow ++;
	}

	///////////////////////////////////////////////////////////////////////////
	if (gpMP3Back)
		UpdateMP3PlayListNextChunk();
}

static void MP3Process(void)
{
	if (gpMP3Back)
	{
		StopMP3IRQ();

		SetClockSpeed(CPU_CLOCK40);

		gpMP3Back = FALSE;
	}
	else
	{
		SetClockSpeed(CPU_CLOCK133);

		if (!PlayMP3PlayList())
			return;

		gpMP3Back = TRUE;
	}
}

static void MP3Stop(void)
{
	if (gpMP3Back)
	{
		StopMP3IRQ();

		SetClockSpeed(CPU_CLOCK40);

		gpMP3Back = FALSE;
	}
}

BOOL TextViewProcess(void)
{
	static BOOL bAccelKey = FALSE;
	static int	nAccelTick = 0;

	int nExKey, nKey;

	//////////////////////////////////////////////////////////////////////////
	nKey = GetKey();
	if (nKey & GPC_VK_FB)
	{
		WaitKey(GPC_VK_FB);

		MP3Stop();
		SaveBookMark();

		DestroyTextView();
		return FALSE;
	}
	if (nKey & GPC_VK_FA)
	{
		WaitKey(GPC_VK_FA);

		if (gpMP3Back)
			SkipMP3PlayList();

		return TRUE;
	}
	else if (nKey & GPC_VK_SELECT)
	{
		WaitKey(GPC_VK_SELECT);

		SetWordWarp();
		return TRUE;
	}
	else if (nKey & GPC_VK_START)
	{
		WaitKey(GPC_VK_START);

		//SetFont();
		MP3Process();
		return TRUE;
	}

	// Accel key
	GpKeyGetEx(&nExKey);
	if (bAccelKey)
	{
		if ((nExKey & GPC_VK_UP) || (nExKey & GPC_VK_DOWN) || (nExKey & GPC_VK_FR) || (nExKey & GPC_VK_FL))
		{
			nKey = nExKey;
		}
		else
		{
			nAccelTick = 0;
			bAccelKey = FALSE;
		}
	}
	else
	{
		if ((nExKey & GPC_VK_UP) || (nExKey & GPC_VK_DOWN) || (nExKey & GPC_VK_FR) || (nExKey & GPC_VK_FL))
		{
			if (nAccelTick)
			{
				if ((GpTickCountGet() - nAccelTick) > MAX_ACCEL_TICK)
					bAccelKey = TRUE;
			}
			else
				nAccelTick = GpTickCountGet();
		}
		else
			nAccelTick = 0;
	}

	if (nKey & GPC_VK_FL)
	{
		MoveLine(-(TEXTVIEW_ROW - 1));
	}
	else if (nKey & GPC_VK_FR)
	{
		MoveLine(TEXTVIEW_ROW - 1);
	}

	if (nKey & GPC_VK_UP)
	{
		MoveLine(-1);
	}
	else if (nKey & GPC_VK_DOWN)
	{
		MoveLine(1);
	}

	return TRUE;
}
