// Viewport.c: implementation of the Viewport function.
//
//////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdio.h>

#include "Viewport.h"

#include "gpstdlib.h"
#include "gpstdio.h"

#include "Textout.h"

#include "Key.h"
#include "PopupMenu.h"

#include "ContextFunc.h"

//
#include "TextView.h"
#include "ImageView.h"

VIEWMODE		gpViewMode = VIEW_FOLDER;
int				gpViewSort = 0;

VIEWPORTINFO	gpViewportInfo[2];
LPVIEWPORTINFO	gpViewport;
int				gpViewportCol;
int				gpViewportRow;

DWORD 	gpTotal, gpUsed, gpFree;

// extern...
extern const unsigned char txt[];
extern const unsigned char fxe[];
extern const unsigned char tga[];
extern const unsigned char bmp[];
extern const unsigned char pic[];
extern const unsigned char mp3[];
extern const unsigned char folder[];
extern const unsigned char file[];

extern const unsigned char open[];
extern const unsigned char cut[];
extern const unsigned char copy[];
extern const unsigned char delete[];

EXTINFO pExtInfo[] =
{
	{ ".",    (LPWORD) folder, VIEW_NONE },
	{ ".",    (LPWORD) file  , VIEW_NONE },
	{ ".TXT", (LPWORD) txt   , VIEW_TEXT },
	{ ".CAP", (LPWORD) txt   , VIEW_TEXT },
	{ ".DOC", (LPWORD) txt   , VIEW_TEXT },
	{ ".FXE", (LPWORD) fxe   , VIEW_EXE },
	{ ".GXB", (LPWORD) fxe   , VIEW_EXE },
	{ ".MP3", (LPWORD) mp3   , VIEW_MP3 },
	{ ".MID", (LPWORD) mp3   , VIEW_MIDI },
	{ ".MOD", (LPWORD) mp3   , VIEW_MOD },
	{ ".SAV", (LPWORD) file  , VIEW_NONE },
	{ ".TGA", (LPWORD) tga   , VIEW_IMAGE },
	{ ".BMP", (LPWORD) bmp   , VIEW_IMAGE },
	{ ".JPG", (LPWORD) pic   , VIEW_IMAGE },
	{ ".PCX", (LPWORD) pic   , VIEW_IMAGE },
	{ NULL, },
};

static int compare_ext(LPDIRENTRY s1, LPDIRENTRY s2);
static int compare_size(LPDIRENTRY s1, LPDIRENTRY s2);
static int compare_name(LPDIRENTRY s1, LPDIRENTRY s2);

static int (*compare)(LPDIRENTRY, LPDIRENTRY) = compare_ext;
static void quicksort(int low, int high, LPDIRENTRY sort);

static void ShowViewport(void)
{
	int nCount, nCol, nRow;
	LPVIEWPORTINFO	pViewport = gpViewportInfo;

	if (gpViewportCol)
	{
		GpFillBlendMask50(2, pViewport->nRow, 155, 218, COLOR_WHITE);
		GpFillBlendMask75(162, pViewport->nRow, 155, 218, COLOR_WHITE);
	}
	else
	{
		GpFillBlendMask75(2, pViewport->nRow, 155, 218, COLOR_WHITE);
		GpFillBlendMask50(162, pViewport->nRow, 155, 218, COLOR_WHITE);
	}

	for (nCol = 0; nCol < 2; nCol ++)
	{
		nRow = MAX_VIEW_ROW;
		if (nRow > pViewport->dwFileCount)
			nRow = pViewport->dwFileCount;

		for (nCount = 0; nCount < nRow; nCount ++)
		{
			int        nSelected = pViewport->nTop + nCount;
			LPDIRENTRY pDirEntry = pViewport->DirEntry + nSelected;

			char szFileName[MAX_PATH_LEN];

			GpPutSprite(pViewport->nCol - 3,
			            pViewport->nRow + 1 + nCount * 12,
			            pExtInfo[pDirEntry->ext].pImage);

			if (pDirEntry->attr & 0x10)
			{
				sprintf(szFileName, "%-12s", pDirEntry->name);
				GpTextOutXY(pViewport->nCol + 10,
				            pViewport->nRow + 2 + nCount * 12,
				            szFileName,
				            0xe0);
			}
			else
			{
				sprintf(szFileName, "%10ldKB", pDirEntry->size / 1024);

				GpTextOutXY(pViewport->nCol + 10,
				            pViewport->nRow + 2 + nCount * 12,
				            pDirEntry->name,
				            0xe0);
				GpTextOutXY(pViewport->nCol + 74,
				            pViewport->nRow + 2 + nCount * 12,
				            szFileName,
				            0xe0);
			}
		}

		pViewport ++;
	}
}

static void ShowCursor(void)
{
	int			nLength;
	char 		szText[MAX_PATH_LEN];
	int 		nSelected = gpViewport->nTop + gpViewport->nSelected;
	LPDIRENTRY	pDirEntry = gpViewport->DirEntry + nSelected;

	GpFillBlendMask75(0, 229, 320, 11, COLOR_MAGENTA);

	GpFill(158, gpViewport->nRow, 2, 217, 0xe0);

	// File count
	sprintf(szText, "%2d/%2ld", nSelected, gpViewport->dwFileCount - 1);
	GpTextOutXY(280, 229, szText, 0x0f);

	// File Info
	sprintf(szText, "%s%-16s", gpViewport->szPath, pDirEntry->name);
	GpTextOutXY(4, 229, szText + 3, 0x0f);	// skip "gp:"

	if (!(pDirEntry->attr & 0x10))
	{
		sprintf(szText, "%12ldBytes", pDirEntry->size);
		GpTextOutXY(186, 229, szText, 0x0f);
	}

	// Volume info
	sprintf(szText, "%ldKB/%ldKB %ldMB", gpFree / 1024, gpUsed / 1024, gpTotal / 1024000);

	nLength = strlen(szText) * 6;
	GpFillBlendMask75(316 - (nLength + 4), 0, nLength + 4, 11, COLOR_BLUE);
	GpTextOutXY(316 - nLength, 0, szText, COLOR_WHITE);

	GpFillBlendMask75(0, 0, 320 - (nLength + 8), 11, COLOR_MAGENTA);
	GpTextOutXY(6, 0, "GP32 File Manager 0.1.2b", COLOR_BLACK);

	// Blending cursor (10-50)
	GpFillBlend(gpViewport->nCol - 5,
	            gpViewport->nRow + 1 + gpViewport->nSelected * 12,
	            153,
	            12,
	            0xf0,
	            gpCursorBlend);
}

static void VK_FA_FUNC(void)
{
	int         nSelected = gpViewport->nTop + gpViewport->nSelected;
	LPDIRENTRY	pDirEntry = gpViewport->DirEntry + nSelected;

	// Folder
	if (pDirEntry->attr & 0x10)
	{
		char* pszPath = gpViewport->szPath;

		if (!strcmp(pDirEntry->name, ".."))
		{
			char* ptr;

			pszPath[strlen(pszPath) - 1] = '\0';		// Cut last character ('\')
			ptr = strrchr(pszPath, '\\');				// Finds last '\'
			if (ptr)
				*(ptr + 1) = '\0';							// Cuts bottom-most folder name
		}
		else
		{
			strcat(gpViewport->szPath, pDirEntry->name);
			strcat(gpViewport->szPath, "\\");
		}

		RefreshViewport(gpViewportCol);

		// Update row
		gpViewportRow = MAX_VIEW_ROW - 1;
		if (gpViewportRow > gpViewport->dwFileCount - 1)
			gpViewportRow = gpViewport->dwFileCount - 1;
	}
	// Viewer
	else
	{
		char szFileName[MAX_PATH_LEN];
		sprintf(szFileName, "%s%s", gpViewport->szPath, pDirEntry->name);

		switch (pExtInfo[pDirEntry->ext].Viewer)
		{
		case VIEW_TEXT:
			if (LoadTextView(szFileName))
			{
				gpViewMode = VIEW_TEXT;
			}
			break;
		case VIEW_IMAGE:
			if (LoadImageView(szFileName))
			{
				gpViewMode = VIEW_IMAGE;
			}
			break;
		}
	}
}
static void VK_FB_FUNC(void)
{
	char* ptr;
	char* pszPath = gpViewport->szPath;

	if (!strcmp(pszPath, "gp:\\"))
		return;

	// Back folder
	pszPath[strlen(pszPath) - 1] = '\0';		// Cut last character ('\')
	ptr = strrchr(pszPath, '\\');				// Finds last '\'
	if (ptr)
		*(ptr + 1) = '\0';						// Cuts bottom-most folder name

	RefreshViewport(gpViewportCol);

	// Update row
	gpViewportRow = MAX_VIEW_ROW - 1;
	if (gpViewportRow > gpViewport->dwFileCount - 1)
		gpViewportRow = gpViewport->dwFileCount - 1;
}

static void SortViewport(void)
{
	gpViewSort ++;
	if (gpViewSort > 2)
		gpViewSort = 0;

	switch (gpViewSort)
	{
	case 0:
		compare = compare_ext;
		break;
	case 1:
		compare = compare_size;
		break;
	case 2:
		compare = compare_name;
		break;
	}

	quicksort(0, gpViewport->dwFileCount - 1, gpViewport->DirEntry);
}

static BOOL KeyProcess()
{
	static BOOL bAccelKey = FALSE;
	static int	nAccelTick = 0;

	int nExKey, nKey;


	GpKeyGetEx(&nExKey);
	if ((nExKey & GPC_VK_START) && (nExKey & GPC_VK_SELECT))
	{
		WaitKey(GPC_VK_START | GPC_VK_SELECT);
		return FALSE;
	}


	//
	nKey = GetKey();
	if (nKey & GPC_VK_FA)
	{
		WaitKey(GPC_VK_FA);

		VK_FA_FUNC();
		return TRUE;
	}
	else if (nKey & GPC_VK_FB)
	{
		WaitKey(GPC_VK_FB);

		VK_FB_FUNC();
		return TRUE;
	}
	else if (nKey & GPC_VK_SELECT)
	{
		WaitKey(GPC_VK_SELECT);

		gpViewMode = VIEW_CONTEXT;
		return TRUE;
	}
	else if (nKey & GPC_VK_FL)
	{
		WaitKey(GPC_VK_FL);

		InitViewport();
		return TRUE;
	}
	else if (nKey & GPC_VK_FR)
	{
		WaitKey(GPC_VK_FR);

		SortViewport();
		return TRUE;
	}

	//
	if (nKey & GPC_VK_LEFT)
	{
		gpViewportCol = 0;
		gpViewport = gpViewportInfo;

		gpViewportRow = MAX_VIEW_ROW - 1;
		if (gpViewportRow > gpViewport->dwFileCount - 1)
			gpViewportRow = gpViewport->dwFileCount - 1;
	}
	else if (nKey & GPC_VK_RIGHT)
	{
		gpViewportCol = 1;
		gpViewport = gpViewportInfo + gpViewportCol;

		gpViewportRow = MAX_VIEW_ROW - 1;
		if (gpViewportRow > gpViewport->dwFileCount - 1)
			gpViewportRow = gpViewport->dwFileCount - 1;
	}

	// Accel key
	if (bAccelKey)
	{
		if ((nExKey & GPC_VK_UP) || (nExKey & GPC_VK_DOWN))
		{
			nKey = nExKey;
		}
		else
		{
			nAccelTick = 0;
			bAccelKey = FALSE;
		}
	}
	else
	{
		if ((nExKey & GPC_VK_UP) || (nExKey & GPC_VK_DOWN))
		{
			if (nAccelTick)
			{
				if ((GpTickCountGet() - nAccelTick) > MAX_ACCEL_TICK)
					bAccelKey = TRUE;
			}
			else
				nAccelTick = GpTickCountGet();
		}
		else
			nAccelTick = 0;
	}

	if (nKey & GPC_VK_UP)
	{
		if (gpViewport->nSelected > 0)
			gpViewport->nSelected --;
		else
		{
			if (gpViewport->nTop > 0)
				gpViewport->nTop --;
			else
			{
				gpViewport->nSelected = gpViewport->dwFileCount - 1;
				if (gpViewport->nSelected < MAX_VIEW_ROW)
					gpViewport->nTop = 0;
				else
				{
					gpViewport->nTop = gpViewport->dwFileCount - MAX_VIEW_ROW;
					gpViewport->nSelected = MAX_VIEW_ROW - 1;
				}
			}
		}
	}
	else if (nKey & GPC_VK_DOWN)
	{
		if (gpViewport->nSelected < gpViewportRow)
			gpViewport->nSelected ++;
		else
		{
			if (gpViewport->nSelected + gpViewport->nTop < (gpViewport->dwFileCount - 1))
				gpViewport->nTop ++;
			else
			{
				gpViewport->nTop = 0;
				gpViewport->nSelected = 0;
			}
		}
	}

	return TRUE;
}

// sort
void splitpath(const char* path, char* name, char* ext)
{
	char* ptr = (char*) path;

	*name = 0;
	*ext  = 0;

	while (*ptr != '.')
	{
		if (!*ptr)
		{
			*name = 0;
			return;
		}

		*name ++ = *ptr ++;
	}
	*name = 0;

	while (*ptr)
	{
		*ext ++ = *ptr ++;
	}
	*ext = 0;
}

static int compare_ext(LPDIRENTRY s1, LPDIRENTRY s2)
{
   	char name1[9], name2[9], ext1[5], ext2[5], path1[14], path2[14];

   	strcpy(path1, s1->name);
   	strcpy(path2, s2->name);

	splitpath(path1, name1, ext1);
	splitpath(path2, name2, ext2);


	// path1
   	if (s1->attr & 0x10)
   		path1[0] = '0';
   	else
   		path1[0] = '1';
	path1[1] = 0;

   	if (strlen(ext1) > 1)
   		strcat(path1, ext1 + 1);
   	strcat(path1, name1);


	// path2
   	if (s2->attr & 0x10)
   		path2[0] = '0';
   	else
   		path2[0] = '1';
	path2[1] = 0;

   	if (strlen(ext2) > 1)
   		strcat(path2, ext2 + 1);
   	strcat(path2, name2);

   	return (strcmp(path1, path2));
}
static int compare_size(LPDIRENTRY s1, LPDIRENTRY s2)
{
   	char path1[26], path2[26];
	char t1 = '1', t2 = '1';

	// path1
   	if (s1->attr & 0x10)
   		t1 = '0';

	// path1
   	if (s2->attr & 0x10)
   		t2 = '0';

   	sprintf(path1, "%11ld%c%s%c", s1->size, t1, s1->name, 0x00);
   	sprintf(path2, "%11ld%c%s%c", s2->size, t2, s2->name, 0x00);

   	return(strcmp(path1, path2));
}
static int compare_name(LPDIRENTRY s1, LPDIRENTRY s2)
{
	char path1[14], path2[14];

	// path1
   	if (s1->attr & 0x10)
   		path1[0] = '0';
   	else
   		path1[0] = '1';
	path1[1] = 0;

   	strcat(path1, s1->name);

	// path2
   	if (s2->attr & 0x10)
   		path2[0] = '0';
   	else
   		path2[0] = '1';
	path2[1] = 0;

   	strcat(path2, s2->name);

   	return (strcmp(path1, path2));
}

static void swaps(LPDIRENTRY s1, LPDIRENTRY s2)
{
   	DIRENTRY temp = *s1;

   	*s1 = *s2; *s2 = temp;
}
static void quicksort(int low, int high, LPDIRENTRY sort)
{
   	int i = 0, j = 0;
   	DIRENTRY temp;

   	if (low < high)
   	{
      	if (high - low == 1)
      	{
        	if (compare(&sort[low], &sort[high]) > 0)
        		swaps(&sort[low], &sort[high]);
      	}
      	else
      	{
        	memcpy(&temp, &sort[high], sizeof(sort[high]));

         	do
         	{
            	i = low; j = high;
            	while ((i < j) && (compare(&sort[i], &temp) <= 0))
            		i ++;

            	while ((j > i) && (compare(&sort[j], &temp) >= 0))
            		j --;

            	if (i < j)
            		swaps(&sort[i], &sort[j]);
         	}
         	while (i < j);

         	swaps(&sort[i], &sort[high]);

         	if ((i - low) < (high - i))
         	{
            	quicksort(low,   i - 1, sort);
            	quicksort(i + 1, high, sort);
         	}
         	else
         	{
            	quicksort(i + 1, high, sort);
            	quicksort(low,   i - 1, sort);
         	}
      	}
   	}
}

static BOOL DirEnumList(const char* lpszPathName, LPDIRENTRY lpDirEntry, LPDWORD lpCount)
{
	int 		nCount;
	ERR_CODE 	result;
	DWORD		dwFileCount = 0;

	DWORD		dwEntryCount;
	GPDIRENTRY	DirEntry[MAX_COUNT_FILE];

	result = GpDirEnumList(lpszPathName, 0, MAX_COUNT_FILE, (GPDIRENTRY *) &DirEntry, &dwEntryCount);
	if ((result != SM_OK) && (result != ERR_EOF))
		return FALSE;

	for (nCount = 0; nCount < dwEntryCount; nCount ++)
	{
		GPFILEATTR	FileAttr;

		char szFileName[MAX_PATH_LEN];
		sprintf(szFileName, "%s%s", lpszPathName, DirEntry[nCount].name);

		result = GpFileAttr(szFileName, &FileAttr);
		if (result != SM_OK)
			return FALSE;

		// folder
		if (FileAttr.attr & 0x10)
		{
			if (!strcmp(DirEntry[nCount].name, "."))
				continue;

			lpDirEntry[dwFileCount].ext = 0;
		}
		else
		{
			int   		nExt = 2;
			LPEXTINFO 	pExt = pExtInfo + nExt;

			char* 		ptr = strchr(DirEntry[nCount].name, '.');

			lpDirEntry[dwFileCount].ext = 1;

			while (pExt->pszExt)
			{
				if (!strcmp(ptr, pExt->pszExt))
				{
					lpDirEntry[dwFileCount].ext = nExt;
					break;
				}

				pExt ++;
				nExt ++;
			}
		}

		lpDirEntry[dwFileCount].attr = FileAttr.attr;
		lpDirEntry[dwFileCount].size = FileAttr.size;

		strcpy(lpDirEntry[dwFileCount].name, DirEntry[nCount].name);

		dwFileCount ++;
	}

	quicksort(0, dwFileCount - 1, lpDirEntry);

	*lpCount = dwFileCount;

	return TRUE;
}

BOOL RefreshViewport(int nViewport)
{
	LPVIEWPORTINFO	pViewport = gpViewportInfo + nViewport;

	pViewport->nTop      = 0;	// Resets top index
	pViewport->nSelected = 0;	// Resets index selected

	SetClockSpeed(CPU_CLOCK133);

	if (!DirEnumList((const char*) pViewport->szPath, (LPDIRENTRY) &pViewport->DirEntry, &pViewport->dwFileCount))
		return FALSE;

	SetClockSpeed(CPU_CLOCK40);

	return TRUE;
}

void InitViewport(void)
{
	char szVolumeName[80] = "gp:\\";

	// Initializes file system
	GpFatInit();

	GpVolumeInfo(szVolumeName, &gpTotal, &gpUsed, &gpFree);

	strcpy(gpViewportInfo[0].szPath, "gp:\\");	// Sets initial path
	strcpy(gpViewportInfo[1].szPath, "gp:\\");	// Sets initial path

	gpViewportInfo[0].nRow = 11;
	gpViewportInfo[0].nCol = 8;
	gpViewportInfo[1].nRow = 11;
	gpViewportInfo[1].nCol = 168;

	RefreshViewport(0);
	RefreshViewport(1);

	gpViewport = gpViewportInfo;

	gpViewportCol  = 0;

	// Update row
	gpViewportRow = MAX_VIEW_ROW - 1;
	if (gpViewportRow > gpViewport->dwFileCount - 1)
		gpViewportRow = gpViewport->dwFileCount - 1;
}

///////////////////////////////////////////////////////////////////////////////
static POPUP pContextMenu[] =
{
	{ "Open",		(LPWORD) open,   OpenFunc,   0 },
	{ "-" },
	{ "Cut",		(LPWORD) cut,    CutFunc,    0 },
	{ "Copy",		(LPWORD) copy,   CopyFunc,   0 },
	{ "Move",		NULL,   		 MoveFunc,   0 },
	{ "-" },
	{ "Delete", 	(LPWORD) delete, DeleteFunc, 0 },
	{ "Rename", 	NULL,            DummyFunc,  0 },
	{ "-" },
	{ "New Folder",	(LPWORD) folder, DummyFunc,  0 },
	{ "-" },
	{ "Properties", NULL,            DummyFunc,  0 },
};
static POPUPMENU pContext =
{
	pContextMenu,
	100, 60,
	90, 150,

	12,
	0,
};

extern VIEWMODE gpOpenViewMode;
extern const unsigned char back[];

BOOL ViewportProcess(void)
{
	LPWORD pBuffer = (LPWORD) back;

	///////////////////////////////////////////////////////////////////////////
	if (gpMP3Back)
		UpdateMP3PlayListNextChunk();

	//GpClear(0xff00);
	GpPutBuffer(pBuffer);

	///////////////////////////////////////////////////////////////////////////
	if (gpMP3Back)
		UpdateMP3PlayListNextChunk();

	if (gpViewMode == VIEW_FOLDER || gpViewMode == VIEW_CONTEXT)
	{
		ShowViewport();

		UpdateCursor();
		ShowCursor();
	}

	///////////////////////////////////////////////////////////////////////////
	if (gpMP3Back)
		UpdateMP3PlayListNextChunk();

	///////////////////////////////////////////////////////////////////////////
	SetKeyStatus();
	switch (gpViewMode)
	{
	case VIEW_FOLDER:
		if (!KeyProcess())
			return FALSE;
		break;

	case VIEW_CONTEXT:
		ShowPopupMenu(&pContext);

		if (!PopupMenuProcess(&pContext))
		{
			char szVolumeName[80] = "gp:\\";
			GpVolumeInfo(szVolumeName, &gpTotal, &gpUsed, &gpFree);

			gpViewMode = gpOpenViewMode;
		}
		break;
	case VIEW_TEXT:
		ShowTextView();

		if (!TextViewProcess())
		{
			if (!gpBookMark)
			{
				int nTop      = gpViewport->nTop;
				int nSelected = gpViewport->nSelected;

				// Initializes file system
				GpFatInit();

				RefreshViewport(gpViewportCol);

				gpViewport->nTop      = nTop;
				gpViewport->nSelected = nSelected;

				if (!strcmp(gpViewportInfo[0].szPath, gpViewportInfo[1].szPath))
				{
					LPVIEWPORTINFO	pViewport = gpViewportInfo + (gpViewportCol ^ 1);

					nTop      = pViewport->nTop;
					nSelected = pViewport->nSelected;

					RefreshViewport(gpViewportCol ^ 1);

					pViewport->nTop      = nTop;
					pViewport->nSelected = nSelected;
				}

				// Update row
				gpViewportRow = MAX_VIEW_ROW - 1;
				if (gpViewportRow > gpViewport->dwFileCount - 1)
					gpViewportRow = gpViewport->dwFileCount - 1;
			}

			gpViewMode = VIEW_FOLDER;
		}
		break;

	case VIEW_IMAGE:
		ShowImageView();

		if (!ImageViewProcess())
			gpViewMode = VIEW_FOLDER;
		break;
	}

	return TRUE;
}
