// Viewport.h: interface for the Viewport func.
//
//////////////////////////////////////////////////////////////////////

#if !defined(_VIEWPORT_H_)
#define _VIEWPORT_H_

#include "gp32lib.h"

#define	MAX_COUNT_FILE	64					// Maximum number of file entries in one folder
#define	MAX_PATH_LEN	255					// Maximum length of full path name
#define MAX_VIEW_ROW	18

// View mode
typedef enum _VIEWMODE
{
	VIEW_NONE,

	VIEW_TEXT,
	VIEW_IMAGE,
	VIEW_EXE,
	VIEW_MP3,
	VIEW_MIDI,
	VIEW_MOD,

	VIEW_FOLDER,
	VIEW_CONTEXT
} VIEWMODE;

// File extension info
typedef struct _EXTINFO
{
	char*		pszExt;
	LPWORD		pImage;

	VIEWMODE	Viewer;
} EXTINFO, *LPEXTINFO;
extern EXTINFO pExtInfo[];

/**********
 * attr
 *	7-6 Reserved. Must be 0's.
 *	5: 1 = ARCHIVE file was modified
 *	4: 1 = DIRECTORY, 0 = file
 *	3: 1 = VOLUME label
 *	2: 1 = SYSTEM file or directory
 *	1: 1 = HIDDEN file or directory
 *	0: 1 = READONLY file
 **********/
typedef struct _DIRENTRY
{
	char 			name[16];

	unsigned long	attr;
	unsigned long 	size;
	unsigned char	ext;
} DIRENTRY, *LPDIRENTRY;

// Viewport info
typedef struct _VIEWPORTINFO
{
	char		szPath[MAX_PATH_LEN + 1];	// Current path (in full path format)
	DIRENTRY	DirEntry[MAX_COUNT_FILE];	// List of files in current path
	DWORD		dwFileCount;				// Number of files in current path
	int			nSelected;					// Index of file entry selected currently
	int			nTop;						// Index of file entry displayed on top in current page
	int			nRow, nCol;
} VIEWPORTINFO, *LPVIEWPORTINFO;

extern VIEWMODE 		gpViewMode;

extern VIEWPORTINFO		gpViewportInfo[2];
extern LPVIEWPORTINFO	gpViewport;
extern int				gpViewportCol;
extern int				gpViewportRow;

void InitViewport(void);
BOOL ViewportProcess(void);

BOOL RefreshViewport(int nViewport);

#endif // !defined(_VIEWPORT_H_)
