#include <string.h>

#include "gp32lib.h"

#include "gpstdlib.h"
#include "gpstdio.h"
#include "gpgraphic.h"

#include "Textout.h"

#define __GP32SDK__

#ifdef __GP32SDK__
static GPDRAWSURFACE	gpDrawSurface[2];		// Priamry & Back SURFACE
#else
static LPWORD			gpLCDAddress[2];
#endif

// Surface address
LPWORD	gpSurface;

char*	gpAppPath;
DWORD   gpFPS;

static BYTE 	gpScreenTableRGB5[32][32];
static BYTE 	gpScreenTableRGB6[64][64];

static void MakeColorTable(void)
{
	int i, j;
	int nMax, nMin;

	// Make screen effect table
	for (i = 0; i < 32; i++)
	{
		for (j = 0; j < 32; j++)
		{
			nMax = max(i, j);
			nMin = min(i, j);

			gpScreenTableRGB5[i][j] = nMax + (31 - nMax) * nMin / 31;
			if (gpScreenTableRGB5[i][j] > 31)
				gpScreenTableRGB5[i][j] = 31;

		}
	}
	for (i = 0; i < 64; i++)
	{
		for (j = 0; j < 64; j++)
		{
			nMax = max(i, j);
			nMin = min(i, j);

			gpScreenTableRGB6[i][j] = nMax + (63 - nMax) * nMin / 63;
			if (gpScreenTableRGB6[i][j] > 63)
				gpScreenTableRGB6[i][j] = 63;
		}
	}
}

BOOL GpInit(void)
{
	// Initializes file system
	GpFatInit();

	GpAppPathSet("gp:\\gpmm\0", 20);

	gpAppPath = GpAppPathGet(NULL);	// Gets the path of current executable file
									// If second parameter is not NULL, length of gpAppPath will be returned
	PcmInit();

	MakeColorTable();

#ifdef __GP32SDK__
	GpGraphicModeSet(16, NULL);

	GpLcdSurfaceGet(&gpDrawSurface[0], 0);
   	GpLcdSurfaceGet(&gpDrawSurface[1], 1);

	// Make gpDraw with primary surface
	GpSurfaceSet(&gpDrawSurface[0]);

	gpSurface = (LPWORD) gpDrawSurface[0].ptbuffer;

	// Option - default LCD is ON
	if (!(GPC_LCD_ON_BIT & GpLcdStatusGet()))
		GpLcdEnable();
#else
	gpLCDAddress[0] = (LPWORD) LCD_FRAMEBUFFER_FRONT;
	gpLCDAddress[1] = (LPWORD) LCD_FRAMEBUFFER_BACK;

	gpSurface = gpLCDAddress[0];

	// 240x320x65536 screen mode, 0 pixel offscreen
	// expects 33MHz HCLK
	rLCDCON1 = (CLKVAL << 8) | (MVAL_USED << 7) | (3 << 5) | (12 << 1) | 1;	// TFT LCD panel,12bpp TFT,ENVID=on
	rLCDCON2 = (VBPD << 24) | ((LCD_PHYSICAL_H - 1) << 14) | (VFPD << 6) | (VSPW);
	rLCDCON3 = (HBPD << 19) | ((LCD_PHYSICAL_W - 1) << 8)  | (HFPD);
	rLCDCON4 = (MVAL << 8)  | (HSPW);
	rLCDCON5 = (1 << 13) | (1 << 10) | (1 << 9) | (1 << 8) | (1 << 2) | 1;	//HSYNC and VSYNC invert its polarity

	//rLCDSADDR1 = (((UINT) gpSurface >> 22) << 21) | M5D((UINT) gpSurface >> 1);
	//rLCDSADDR2 = M5D(((UINT) gpSurface + (LCD_PHYSICAL_BUFFER * 2)) >> 1);
	//rLCDSADDR3 = (LCD_PHYSICAL_W / 1);

	rLCDSADDR1 = (((UINT) gpSurface & 0x0fc00000) >> 1) | (((UINT) gpSurface & 0x3ffffe) >> 1);
	rLCDSADDR2 = (((UINT) gpSurface & 0x003ffffe) >> 1) + (LCD_PHYSICAL_BUFFER);
	rLCDSADDR3 = (LCD_PHYSICAL_W);

	rTPAL = 0;	// Disable Temporary Palette

	// Enable LCD
	rLCDCON1 |= 1; // ENVID=ON
#endif

	return TRUE;
}

void GpClose(void)
{
	PcmClose();

	GpAppExit();
}

void GpPresent(void)
{
	static int nSurface = 0;

#ifdef __GP32SDK__
	GpSurfaceFlip(&gpDrawSurface[nSurface]);

	nSurface ^= 0x01;
	gpSurface = (LPWORD) gpDrawSurface[nSurface].ptbuffer;
#else
	// Wait for specified LCD line. Note that the LCD must be enabled
	while (rLCDCON1 >> 18);
	//while ((rLCDCON5 >> 19) == 0x02);

	rLCDSADDR1 = (((UINT) gpSurface & 0x0fc00000) >> 1) | (((UINT) gpSurface & 0x3ffffe) >> 1);
	rLCDSADDR2 = (((UINT) gpSurface & 0x003ffffe) >> 1) + (LCD_PHYSICAL_BUFFER);

	//rLCDSADDR1 = (((UINT) gpSurface >> 22) << 21) | M5D((UINT) gpSurface >> 1);
	//rLCDSADDR2 = M5D(((UINT) gpSurface + (LCD_PHYSICAL_BUFFER * 2)) >> 1);

	nSurface ^= 0x01;
	gpSurface = gpLCDAddress[nSurface];
#endif
}

///////////////////////////////////////////////////////////////////////////////
void GpUpdateFPS(void)
{
	static DWORD dwLastTickCount = 0L;
    static DWORD dwFrames = 0;

	DWORD dwTickCount = GpTickCountGet();
    ++ dwFrames;

    if (dwTickCount - dwLastTickCount > 1000)
    {
        gpFPS = dwFrames / ((dwTickCount - dwLastTickCount) / 1000);
        dwFrames = 0;

        dwLastTickCount = dwTickCount;
	}
}

//
// DMA0 memory to memory copy routine..
// Note: size = number or words to copy.
//
// This DMA copy routine does not use interrupts to inform when the
// DMA has finished.. which is kinda lame. The caller must use the
// 'DMA0WaitEnd()' function to check/busywait to see when DMA has finished
// the data transfer.
//
void DMA0CopyStart(LPWORD lpSource, LPWORD lpDest, DWORD dwLength)
{
	// Stop DMA0
	rDMASKTRIG0 |= 0x04;

	// Setup source...
	//   select system bus (AHB)
	//   select address increment
	rDISRC0 = (long) lpSource;

	// Setup destination...
	//   select system bus (AHB)
	//   select address increment
	rDIDST0 = (long) lpDest;

	// DMA0 con..
	rDCON0 = 	(0 << 30) | 		// demand mode
				(1 << 29) |			// DREQ & DACK are synched to HCLK (AHB)
				(0 << 28) |			// disable DMA0 int
				(0 << 27) |			// unit transfer
				(1 << 26) |			// whole servmode
				(0x00 << 24) |		// HWSRCSEL ?? not used
				(0 << 23) |			// sw request mode
				(1 << 22) |			// auto reload off
				(0x02 << 20) |		// data size == word
				(dwLength >> 1);	// count..

	// start it
	rDMASKTRIG0 = 0x03;
}

//
void DMA0WaitEnd(void)
{
	// poll for end..
	while ((rDSTAT0 >> 20) & 0x03);
}

///////////////////////////////////////////////////////////////////////////////
void GpClear(WORD wColor)
{
	int   nLength = LCD_PHYSICAL_BUFFER / 2;
	DWORD dwColor = MAKEDWORD(wColor, wColor);

	LPDWORD pSurface = (LPDWORD) gpSurface;
	while (nLength --)
		*pSurface ++ = dwColor;
}

void GpPutBuffer(LPWORD lpBuffer)
{
	int nLength = LCD_PHYSICAL_BUFFER / 2;

	LPDWORD pBuffer  = (LPDWORD) lpBuffer;
	LPDWORD pSurface = (LPDWORD) gpSurface;
	while (nLength --)
		*pSurface ++ = *pBuffer ++;
}

void GpPutOffset(int x, int y, LPWORD lpBuffer)
{
	int nRow = LCD_PHYSICAL_H;
	int nWidth = LCD_PHYSICAL_W / 2;

	LPWORD pBuffer   = (LPWORD) (lpBuffer + ((LCD_PHYSICAL_W + 16) * (16 - x) + y));
	LPDWORD pSurface = (LPDWORD) gpSurface;

	while (nRow --)
	{
		int nCol = nWidth;
		while (nCol --)
		{
			*pSurface ++ = MAKEDWORD(*pBuffer, *(pBuffer + 1));
			pBuffer += 2;
		}

		pBuffer += 16;
	}
}

void GpHLine(int x, int y, int nWidth, WORD wColor)
{
	int nRow = nWidth;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y);
	while (nRow --)
	{
		*pSurface = wColor;
		pSurface += LCD_PHYSICAL_W;
	}
}
void GpVLine(int x, int y, int nHeight, WORD wColor)
{
	int nCount = nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);
	while (nCount --)
		*pSurface ++ = wColor;
}
void GpFill(int x, int y, int nWidth, int nHeight, WORD wColor)
{
	int nRow = nWidth;
	int nSkipCol = LCD_PHYSICAL_W - nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);

	while (nRow --)
	{
		int nCol = nHeight;
		while (nCol --)
			*pSurface ++ = wColor;

		pSurface += nSkipCol;
	}
}

void GpHLineBlend(int x, int y, int nWidth, WORD wColor, int nBlend)
{
	int nRow = nWidth;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y);
	while (nRow --)
	{
		BYTE r, g, b;
		WORD color;

		r = (GETR(*pSurface) * (100 - nBlend) + GETR(wColor) * nBlend) / 100;
    	g = (GETG(*pSurface) * (100 - nBlend) + GETG(wColor) * nBlend) / 100;
    	b = (GETB(*pSurface) * (100 - nBlend) + GETB(wColor) * nBlend) / 100;

    	color = MAKERGB2(r, g, b);
		*pSurface = color;

		pSurface += LCD_PHYSICAL_W;
	}
}
void GpVLineBlend(int x, int y, int nHeight, WORD wColor, int nBlend)
{
	int nCount = nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);
	while (nCount --)
	{
		BYTE r, g, b;
		WORD color;

		r = (GETR(*pSurface) * (100 - nBlend) + GETR(wColor) * nBlend) / 100;
    	g = (GETG(*pSurface) * (100 - nBlend) + GETG(wColor) * nBlend) / 100;
    	b = (GETB(*pSurface) * (100 - nBlend) + GETB(wColor) * nBlend) / 100;

    	color = MAKERGB2(r, g, b);
		*pSurface ++ = color;
	}
}
void GpFillBlend(int x, int y, int nWidth, int nHeight, WORD wColor, int nBlend)
{
	int nRow = nWidth;
	int nSkipCol = LCD_PHYSICAL_W - nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);

	while (nRow --)
	{
		int nCol = nHeight;
		while (nCol --)
		{
			BYTE r, g, b;
			WORD color;

			r = (GETR(*pSurface) * (100 - nBlend) + GETR(wColor) * nBlend) / 100;
        	g = (GETG(*pSurface) * (100 - nBlend) + GETG(wColor) * nBlend) / 100;
        	b = (GETB(*pSurface) * (100 - nBlend) + GETB(wColor) * nBlend) / 100;

        	color = MAKERGB2(r, g, b);
			*pSurface ++ = color;
		}

		pSurface += nSkipCol;
	}
}

void GpHLineBlendMask(int x, int y, int nWidth, WORD wColor)
{
	int nRow = nWidth;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y);
	while (nRow --)
	{
		*pSurface = BLEND75(wColor, *pSurface);

		pSurface += LCD_PHYSICAL_W;
	}
}
void GpVLineBlendMask(int x, int y, int nHeight, WORD wColor)
{
	int nCount = nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);
	while (nCount --)
	{
		*pSurface = BLEND75(wColor, *pSurface);
		pSurface ++;
	}
}
void GpFillBlendMask25(int x, int y, int nWidth, int nHeight, WORD wColor)
{
	int nRow = nWidth;
	int nSkipCol = LCD_PHYSICAL_W - nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);

	while (nRow --)
	{
		int nCol = nHeight;
		while (nCol --)
		{
			*pSurface = BLEND25(wColor, *pSurface);
			pSurface ++;
		}

		pSurface += nSkipCol;
	}
}
void GpFillBlendMask50(int x, int y, int nWidth, int nHeight, WORD wColor)
{
	int nRow = nWidth;
	int nSkipCol = LCD_PHYSICAL_W - nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);

	while (nRow --)
	{
		int nCol = nHeight;
		while (nCol --)
		{
			*pSurface = BLEND50(wColor, *pSurface);
			pSurface ++;
		}

		pSurface += nSkipCol;
	}
}
void GpFillBlendMask75(int x, int y, int nWidth, int nHeight, WORD wColor)
{
	int nRow = nWidth;
	int nSkipCol = LCD_PHYSICAL_W - nHeight;

	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nHeight - 1);

	while (nRow --)
	{
		int nCol = nHeight;
		while (nCol --)
		{
			*pSurface = BLEND75(wColor, *pSurface);
			pSurface ++;
		}

		pSurface += nSkipCol;
	}
}

void GpPutImage(int x, int y, int nWidth, int nHeight, LPWORD lpBuffer)
{
	int nRow = nHeight;
	int nSkipCol = LCD_PHYSICAL_W - nWidth;

	LPWORD pBuffer  = lpBuffer;
	LPWORD pSurface = LCD_GET_ADDRESS(x, y + nWidth - 1);

	while (nRow --)
	{
		int nCol = nWidth;
		while (nCol --)
			*pSurface ++ = *pBuffer ++;

		pSurface += nSkipCol;
		/*
		DMA0CopyStart(pBuffer, pSurface, nWidth);
		DMA0WaitEnd();

		pBuffer  += nWidth;
		pSurface += LCD_PHYSICAL_W;
		*/
	}
}

void GpPutSprite(int x, int y, LPWORD lpBuffer)
{
	int nRow;

	WORD wWidth, wHeight, wIndex;
	WORD wSkipByte, wPutByte;

	LPWORD pBuffer  = lpBuffer;
	LPWORD pSurface;

	wWidth  = *pBuffer ++;
	wHeight = *pBuffer ++;

	pSurface = LCD_GET_ADDRESS(x, y + wWidth - 1);

	nRow = wHeight;
	while (nRow --)
	{
		LPWORD pLineSurface = pSurface;

		wIndex = *pBuffer ++;
		while (wIndex --)
		{
			wSkipByte = *pBuffer ++;
			wPutByte  = *pBuffer ++;

			pLineSurface += wSkipByte;
			if (wPutByte != 0)
			{
				while (wPutByte --)
					*pLineSurface ++ = *pBuffer ++;
			}
		}

		pSurface += LCD_PHYSICAL_W;
	}
}

void GpPutScreen(int x, int y, LPWORD lpBuffer)
{
	int nRow;

	WORD wWidth, wHeight, wIndex;
	WORD wSkipByte, wPutByte;

	LPWORD pBuffer  = lpBuffer;
	LPWORD pSurface;

	wWidth  = *pBuffer ++;
	wHeight = *pBuffer ++;

	pSurface = LCD_GET_ADDRESS(x, y + wWidth - 1);

	nRow = wHeight;
	while (nRow --)
	{
		LPWORD pLineSurface = pSurface;

		wIndex = *pBuffer ++;
		while (wIndex --)
		{
			wSkipByte = *pBuffer ++;
			wPutByte  = *pBuffer ++;

			pLineSurface += wSkipByte;
			if (wPutByte != 0)
			{
				while (wPutByte --)
				{
					BYTE r, g, b;

					r = gpScreenTableRGB5[(*pBuffer & 0xF800) >> 11][(*pLineSurface & 0xF800) >> 11];
					g = gpScreenTableRGB6[(*pBuffer & 0x07e0) >> 5][(*pLineSurface & 0x07e0) >> 5];
					b = gpScreenTableRGB5[(*pBuffer & 0x001f)][*pLineSurface & 0x001f];

					*pLineSurface ++ = (WORD) ((r << 11) | (g << 5) | b);

					pBuffer ++;
				}
			}
		}

		pSurface += LCD_PHYSICAL_W;
	}
}

void GpPutBlend(int x, int y, LPWORD lpBuffer, int nBlend)
{
	int nRow;

	WORD wWidth, wHeight, wIndex;
	WORD wSkipByte, wPutByte;

	LPWORD pBuffer  = lpBuffer;
	LPWORD pSurface;

	wWidth  = *pBuffer ++;
	wHeight = *pBuffer ++;

	pSurface = LCD_GET_ADDRESS(x, y + wWidth - 1);

	nRow = wHeight;
	while (nRow --)
	{
		LPWORD pLineSurface = pSurface;

		wIndex = *pBuffer ++;
		while (wIndex --)
		{
			wSkipByte = *pBuffer ++;
			wPutByte  = *pBuffer ++;

			pLineSurface += wSkipByte;
			if (wPutByte != 0)
			{
				while (wPutByte --)
				{
					BYTE r, g, b;
					WORD color;

					r = (GETR(*pLineSurface) * (100 - nBlend) + GETR(*pBuffer) * nBlend) / 100;
		        	g = (GETG(*pLineSurface) * (100 - nBlend) + GETG(*pBuffer) * nBlend) / 100;
		        	b = (GETB(*pLineSurface) * (100 - nBlend) + GETB(*pBuffer) * nBlend) / 100;

		        	color = MAKERGB2(r, g, b);

					*pLineSurface = color;

					pLineSurface ++;
					pBuffer ++;
				}
			}
		}

		pSurface += LCD_PHYSICAL_W;
	}
}

void GpPutBlendMask(int x, int y, LPWORD lpBuffer)
{
	int nRow;

	WORD wWidth, wHeight, wIndex;
	WORD wSkipByte, wPutByte;

	LPWORD pBuffer  = lpBuffer;
	LPWORD pSurface;

	wWidth  = *pBuffer ++;
	wHeight = *pBuffer ++;

	pSurface = LCD_GET_ADDRESS(x, y + wWidth - 1);

	nRow = wHeight;
	while (nRow --)
	{
		LPWORD pLineSurface = pSurface;

		wIndex = *pBuffer ++;
		while (wIndex --)
		{
			wSkipByte = *pBuffer ++;
			wPutByte  = *pBuffer ++;

			pLineSurface += wSkipByte;
			if (wPutByte != 0)
			{
				while (wPutByte --)
				{
					*pLineSurface = (WORD) BLEND50(*pBuffer, *pLineSurface);

					pLineSurface ++;
					pBuffer ++;
				}
			}
		}

		pSurface += LCD_PHYSICAL_W;
	}
}

LPBYTE GpLoadFromFile(const char* lpszFileName, LPDWORD lpLength)
{
	ERR_CODE 	result;
	F_HANDLE 	handle;

	LPBYTE 		pBuffer = NULL;

	DWORD dwSize, dwDummy;
	GpFileGetSize(lpszFileName, &dwSize);

	result = GpFileOpen(lpszFileName, OPEN_R, &handle);
	if (result != SM_OK)
		return NULL;

	pBuffer = (LPBYTE) gp_mem_func.malloc(dwSize);
	if (!pBuffer)
	{
		GpFileClose(handle);
		return NULL;
	}

	GpFileRead(handle, pBuffer, dwSize, &dwDummy);
	GpFileClose(handle);

	*lpLength = dwSize;

	return pBuffer;
}

void GpDelay(DWORD dwTime)
{
	DWORD dwTick = GpTickCountGet();

	// synchronizes frame rate with about 1000 msec
	while ((GpTickCountGet() - dwTick) < dwTime);
}

