#ifndef __GP3D_H__
#define __GP3D_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "gp32.h"
#include "gpdef.h"


typedef int BOOL;
#define TRUE	1
#define FALSE	0

typedef unsigned char	BYTE,	*LPBYTE;
typedef unsigned short  WORD,	*LPWORD;
typedef unsigned long	DWORD,	*LPDWORD;
typedef unsigned int    UINT,   *LPUINT;

extern LPWORD 	gpSurface;

extern char*	gpAppPath;
extern DWORD    gpFPS;

#define MAKEDWORD(lo, hi)	((DWORD) (((hi) << 16 ) | (lo)))

#define max(a, b)    		(((a) > (b)) ? (a) : (b))
#define min(a, b)    		(((a) < (b)) ? (a) : (b))


#define MVAL		(13)
#define MVAL_USED 	(0)
#define M5D(n) 		((n) & 0x1fffff)

#define CLKVAL		(5)
    // 66.54hz @67.71Mhz
    //VSYNC,HSYNC should be inverted
    //HBPD=7VCLK,HFPD=3VCLK,HSPW=4VCLK
    //VBPD=2HSYNC,VFPD=3HSYNC,VSPW=2HSYNC

// Timing parameter for TFT 240320(SAMSUNG 3.5 inch)
#define VBPD			((2 - 1) & 0xff)
#define VFPD			((3 - 1) & 0xff)
#define VSPW			((2 - 1) & 0x3f)
#define HBPD			((7 - 1) & 0x7f)
#define HFPD			((3 - 1) & 0xff)
#define HSPW			((4 - 1) & 0xff)

// TFT 240320
//    0x0C7B4000             - Bitmap memory area starts here
//    0x0C7FF800             - Bitmap memory area ends here
//	  	FRONT 	0x0C7B4000
//		BACK	0x0C7D9800	FRONT + 320 * 240 * 2
#define LCD_FRAMEBUFFER_FRONT	0x0C7B4000
#define LCD_FRAMEBUFFER_BACK	0x0C7D9800

#define LCD_PHYSICAL_W			(240)
#define LCD_PHYSICAL_H			(320)
#define LCD_PHYSICAL_BUFFER		(LCD_PHYSICAL_W * LCD_PHYSICAL_H)

///////////////////////////////////////////////////////////////////////////////
#define LCD_PUT_PIXEL(x, y, color)	\
	gpSurface[(x) * LCD_PHYSICAL_W + LCD_PHYSICAL_W - (y) - 1] = color;

#define LCD_GET_ADDRESS(x, y)		\
	&gpSurface[(x) * LCD_PHYSICAL_W + LCD_PHYSICAL_W - (y) - 1];

// RGB 16bit
#define RGB5(value)			((WORD) ((value >> 3) & 0x001F))
#define MAKERGB(r, g, b)	(RGB5(r) << 11 | RGB5(g) << 6 | RGB5(b) << 1 | 0x00)
#define MAKERGB2(r, g, b)	((r) << 11 | (g) << 6 | (b) << 1 | 0x00)

#define GETR(value)			(((value) >> 11) & 0x001f)
#define GETG(value)			(((value) >> 6)  & 0x001f)
#define GETB(value)			(((value) >> 1)  & 0x001f)

#define RGBHIGH(r, g, b)	(((r) & 0xf8) | ((g) >> 5))
#define RGBLOW(r, g, b)		(((g) << 3) & 0xc0 | ((b) >> 2) & 0x3e)
//
#define BLEND50(a, b)		(((a) >> 1) & 0x7BDE) + (((b) >> 1) & 0x7BDE)
#define BLEND25(a, b)		(((a) >> 2) & 0x39CE) + (((b) >> 1) & 0x7BDE) + (((b) >> 2) & 0x39CE)
#define BLEND75(a, b)		BLEND25((b), (a))
//
#define COLOR_WHITE			0xffff
#define COLOR_BLACK			0x0000
#define COLOR_RED			0xf800
#define COLOR_GREEN			0x07e0
#define COLOR_BLUE			0x003e
#define COLOR_MAGENTA		0xf83e
#define COLOR_CYAN			0x07fe
#define COLOR_YELLOW		0xffc0
#define COLOR_ORANGE		0xfc00
#define COLOR_VIOLET		COLOR_MAGENTA

//
BOOL GpInit(void);
void GpClose(void);

void GpPresent(void);

void GpUpdateFPS(void);

void GpClear(WORD wColor);

void GpHLine(int x, int y, int nWidth, WORD wColor);
void GpVLine(int x, int y, int nHeight, WORD wColor);
void GpFill(int x, int y, int nWidth, int nHeight, WORD wColor);

void GpHLineBlend(int x, int y, int nWidth, WORD wColor, int nBlend);
void GpVLineBlend(int x, int y, int nHeight, WORD wColor, int nBlend);
void GpFillBlend(int x, int y, int nWidth, int nHeight, WORD wColor, int nBlend);

void GpHLineBlendMask(int x, int y, int nWidth, WORD wColor);
void GpVLineBlendMask(int x, int y, int nHeight, WORD wColor);

void GpFillBlendMask25(int x, int y, int nWidth, int nHeight, WORD wColor);
void GpFillBlendMask50(int x, int y, int nWidth, int nHeight, WORD wColor);
void GpFillBlendMask75(int x, int y, int nWidth, int nHeight, WORD wColor);

void GpPutBuffer(LPWORD lpBuffer);
void GpPutOffset(int x, int y, LPWORD lpBuffer);

void GpPutImage(int x, int y, int nWidth, int nHeight, LPWORD lpBuffer);
void GpPutSprite(int x, int y, LPWORD lpBuffer);
void GpPutScreen(int x, int y, LPWORD lpBuffer);
void GpPutBlend(int x, int y, LPWORD lpBuffer, int nBlend);
void GpPutBlendMask(int x, int y, LPWORD lpBuffer);

// File
LPBYTE GpLoadFromFile(const char* lpszFileName, LPDWORD lpLength);

//
void GpDelay(DWORD dwTime);

// Sound API
#define CPU_CLOCK40		0
#define CPU_CLOCK67		1
#define CPU_CLOCK133	2

BOOL PcmInit(void);
void PcmClose(void);

void PlayPcm(unsigned char* src, int size, int rpt);

//
#define CHUNK_LENGTH	0x1200

extern DWORD	dwPcmChunkOffset, dwPcmChunkLength;
extern LPBYTE	pPcmChunkBuffer;

void DMA2_ISR_Install(void);
void DMA2_ISR_Uninstall(void);

void PlayPcmNextChunk(void);

BOOL CreatePcmChunk(void);
void DestroyPcmChunk(void);

// PCM
BOOL PlayPcmIRQ(const char* lpszFileName);
void StopPcmIRQ(void);


// MP3
BOOL PlayMP3IRQ(const char* lpszFileName);
void StopMP3IRQ(void);

void UpdateMP3NextChunk(void);

//
BOOL PlayMP3PlayList(void);

void UpdateMP3PlayListNextChunk(void);
BOOL SkipMP3PlayList(void);

char* GetMP3PlayList(void);

BOOL MP3ToPCM(const char* lpszFileName, const char* lpszPCMName);

// CPU
void SetClockSpeed(int nClockSpeed);

#ifdef __cplusplus
}
#endif

#endif
