#include <string.h>
#include <stdio.h>

#include "gp32lib.h"

#include "gpmem.h"
#include "gpstdio.h"
#include "gpstdlib.h"

#include "libmad/stream.h"
#include "libmad/synth.h"
#include "libmad/frame.h"

#include "libmad/audio.h"

static F_HANDLE	fHandle;
static DWORD	dwFileLength, dwFileOffset;

static LPBYTE	pPcmOffset;
static DWORD	dwPcmOffset;

static struct mad_stream 	stream;
static struct mad_frame 	frame;
static struct mad_synth 	synth;

#define MPEG_BUFSZ	40000	/* 2.5 s at 128 kbps; 1 s at 320 kbps */

static LPBYTE 		pMP3ChunkBuffer;
static DWORD		dwMP3ChunkLength;

/*
 * NAME:	mono_filter()
 * DESCRIPTION:	transform stereo frame to mono
 */
void mono_filter(struct mad_frame *frame)
{
  	if (frame->header.mode != MAD_MODE_SINGLE_CHANNEL)
  	{
    	unsigned int ns, s, sb;
    	mad_fixed_t left, right;

    	ns = MAD_NSBSAMPLES(&frame->header);

    	for (s = 0; s < ns; ++s)
    	{
      		for (sb = 0; sb < 32; ++sb)
      		{
				left  = frame->sbsample[0][s][sb];
				right = frame->sbsample[1][s][sb];

				frame->sbsample[0][s][sb] = (left + right) / 2;
      		}
    	}

    	frame->header.mode = MAD_MODE_SINGLE_CHANNEL;
  	}
}

/*
 * NAME:	experimental_filter()
 * DESCRIPTION:	experimental filter
 */
void experimental_filter(struct mad_frame *frame)
{
  	if (frame->header.mode == MAD_MODE_STEREO || frame->header.mode == MAD_MODE_JOINT_STEREO)
  	{
    	unsigned int ns, s, sb;

    	ns = MAD_NSBSAMPLES(&frame->header);

    	// enhance stereo separation
    	for (s = 0; s < ns; ++s)
    	{
      		for (sb = 0; sb < 32; ++sb)
      		{
				mad_fixed_t left, right;

				left  = frame->sbsample[0][s][sb];
				right = frame->sbsample[1][s][sb];

				frame->sbsample[0][s][sb] -= right / 4;
				frame->sbsample[1][s][sb] -= left  / 4;
      		}
    	}
  	}
}

void UpdateMP3NextChunk(void)
{
	static DWORD dwChunkOffset = 0;

	DWORD dwLength;
	mad_fixed_t const *ch1, *ch2;

	if (dwPcmChunkOffset == dwChunkOffset)
		return;

	dwChunkOffset = dwPcmChunkOffset;

	///////////////////////////////////////////////////////////////////////////
	if (stream.next_frame)
		memmove(pMP3ChunkBuffer,
		        stream.next_frame,
		        dwMP3ChunkLength = &pMP3ChunkBuffer[dwMP3ChunkLength] - stream.next_frame);

	if (dwFileOffset > (dwFileLength - 1))
	{
		GpFileSeek(fHandle, FROM_BEGIN, 0, NULL);

		dwFileOffset = 0;
	}

	GpFileRead(fHandle, pMP3ChunkBuffer + dwMP3ChunkLength, MPEG_BUFSZ - dwMP3ChunkLength, &dwLength);
	dwFileOffset += dwLength;

	mad_stream_buffer(&stream, pMP3ChunkBuffer, dwMP3ChunkLength += dwLength);

	mad_frame_decode(&frame, &stream);

	//mono_filter(&frame);
	//experimental_filter(&frame);

	mad_synth_frame(&synth, &frame);

	///////////////////////////////////////////////////////////////////////////
	ch1 = synth.pcm.samples[0];
  	ch2 = synth.pcm.samples[1];

	if (synth.pcm.channels == 1)
		ch2 = ch1;

  	dwLength = audio_pcm_s16le(pPcmOffset, synth.pcm.length, ch1, ch2);

	dwPcmOffset += dwLength;
	if (dwPcmOffset > (dwPcmChunkLength - 1))
	{
		dwPcmOffset = 0;
		pPcmOffset  = pPcmChunkBuffer;

		return;
	}

	pPcmOffset += dwLength;
}

BOOL PlayMP3IRQ(const char* lpszFileName)
{
	ERR_CODE result;

	// Get file size
	result = GpFileGetSize(lpszFileName, &dwFileLength);
	if (result != SM_OK)
		return FALSE;

	// Open file
	result = GpFileOpen(lpszFileName, OPEN_R, &fHandle);
	if (result != SM_OK)
		return FALSE;

	//
	if (!CreatePcmChunk())
		return FALSE;

	dwPcmOffset = 0;
	pPcmOffset  = pPcmChunkBuffer;

	dwFileOffset = 0;

	//
	pMP3ChunkBuffer = (LPBYTE) gm_malloc(MPEG_BUFSZ);
	if (!pMP3ChunkBuffer)
		return FALSE;

	dwMP3ChunkLength = 0;

	// init mad
	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);

	audio_init();

	// Install dma2-irq
    DMA2_ISR_Install();

	PlayPcmNextChunk();

	//GpTimerOptSet(0, 50, 0, UpdateMP3NextChunk);
	//GpTimerSet(0);

	return TRUE;
}

void StopMP3IRQ(void)
{
	//GpTimerKill(0);

	// Uninstall dma2-irq
    DMA2_ISR_Uninstall();

	DestroyPcmChunk();

 	gm_free(pMP3ChunkBuffer);

	mad_synth_finish(&synth);
	mad_frame_finish(&frame);
	mad_stream_finish(&stream);

	if (fHandle)
	{
    	GpFileClose(fHandle);
    	fHandle = 0;
    }
}

///////////////////////////////////////////////////////////////////////////////
#define	MAX_COUNT_FILE	64					// Maximum number of file entries in one folder
#define	MAX_PATH_LEN	255					// Maximum length of full path name

static GPDIRENTRY	MP3FileList[MAX_COUNT_FILE];		// List of files in current path
static DWORD		dwMP3FileCount, dwMP3FileCurrent;	// Number of files in current

static char*		MP3Path = "gp:\\MP3\\";

void UpdateMP3PlayListNextChunk(void)
{
	static DWORD dwChunkOffset = 0;

	DWORD dwLength;
	mad_fixed_t const *ch1, *ch2;

	if (dwPcmChunkOffset == dwChunkOffset)
		return;

	dwChunkOffset = dwPcmChunkOffset;

	///////////////////////////////////////////////////////////////////////////
	if (stream.next_frame)
		memmove(pMP3ChunkBuffer,
		        stream.next_frame,
		        dwMP3ChunkLength = &pMP3ChunkBuffer[dwMP3ChunkLength] - stream.next_frame);

	if (dwFileOffset > (dwFileLength - 1))
	{
		char szFileName[MAX_PATH_LEN];

		if (fHandle)
		{
	    	GpFileClose(fHandle);
	    	fHandle = 0;
	    }

		dwMP3FileCurrent ++;
		if (dwMP3FileCurrent > (dwMP3FileCount - 1))
			dwMP3FileCurrent = 2;

		//
		sprintf(szFileName, "%s%s", MP3Path, MP3FileList[dwMP3FileCurrent].name);

		// Get file size
		GpFileGetSize(szFileName, &dwFileLength);

		// Open file
		GpFileOpen(szFileName, OPEN_R, &fHandle);

		dwFileOffset = 0;

		dwMP3ChunkLength = 0;

		GpDelay(500);
	}

	GpFileRead(fHandle, pMP3ChunkBuffer + dwMP3ChunkLength, MPEG_BUFSZ - dwMP3ChunkLength, &dwLength);
	dwFileOffset += dwLength;

	mad_stream_buffer(&stream, pMP3ChunkBuffer, dwMP3ChunkLength += dwLength);

	mad_frame_decode(&frame, &stream);

	//mono_filter(&frame);
	//experimental_filter(&frame);

	mad_synth_frame(&synth, &frame);

	///////////////////////////////////////////////////////////////////////////
	ch1 = synth.pcm.samples[0];
  	ch2 = synth.pcm.samples[1];

	if (synth.pcm.channels == 1)
		ch2 = ch1;

  	dwLength = audio_pcm_s16le(pPcmOffset, synth.pcm.length, ch1, ch2);

	dwPcmOffset += dwLength;
	if (dwPcmOffset > (dwPcmChunkLength - 1))
	{
		dwPcmOffset = 0;
		pPcmOffset  = pPcmChunkBuffer;

		return;
	}

	pPcmOffset += dwLength;
}

BOOL SkipMP3PlayList(void)
{
	ERR_CODE result;
	char szFileName[MAX_PATH_LEN];

	StopMP3IRQ();

	dwMP3FileCurrent ++;
	if (dwMP3FileCurrent > (dwMP3FileCount - 1))
		dwMP3FileCurrent = 2;

	//
	sprintf(szFileName, "%s%s", MP3Path, MP3FileList[dwMP3FileCurrent].name);

	// Get file size
	result = GpFileGetSize(szFileName, &dwFileLength);
	if (result != SM_OK)
		return FALSE;

	// Open file
	result = GpFileOpen(szFileName, OPEN_R, &fHandle);
	if (result != SM_OK)
		return FALSE;

	dwFileOffset = 0;

	//
	if (!CreatePcmChunk())
		return FALSE;

	dwPcmOffset = 0;
	pPcmOffset  = pPcmChunkBuffer;

	//
	pMP3ChunkBuffer = (LPBYTE) gm_malloc(MPEG_BUFSZ);
	if (!pMP3ChunkBuffer)
		return FALSE;

	dwMP3ChunkLength = 0;

	// init mad
	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);

	audio_init();

	// Install dma2-irq
    DMA2_ISR_Install();

	UpdateMP3PlayListNextChunk();

	PlayPcmNextChunk();
	GpDelay(500);

	return TRUE;
}

BOOL PlayMP3PlayList(void)
{
	ERR_CODE result;
	char szFileName[MAX_PATH_LEN];

	dwMP3FileCurrent = 2;

	// Initializes file system
	GpFatInit();

	// get mp3 list
	result = GpDirEnumList((const char*) MP3Path,
	                        0,
	                        MAX_COUNT_FILE,
	                        (GPDIRENTRY *) &MP3FileList,
	                        &dwMP3FileCount);
	if ((result != SM_OK) && (result != ERR_EOF))
		return FALSE;

	// open first mp3
	sprintf(szFileName, "%s%s", MP3Path, MP3FileList[dwMP3FileCurrent].name);

	// Get file size
	result = GpFileGetSize(szFileName, &dwFileLength);
	if (result != SM_OK)
		return FALSE;

	// Open file
	result = GpFileOpen(szFileName, OPEN_R, &fHandle);
	if (result != SM_OK)
		return FALSE;

	//
	if (!CreatePcmChunk())
		return FALSE;

	dwPcmOffset = 0;
	pPcmOffset  = pPcmChunkBuffer;

	dwFileOffset = 0;

	//
	pMP3ChunkBuffer = (LPBYTE) gm_malloc(MPEG_BUFSZ);
	if (!pMP3ChunkBuffer)
		return FALSE;

	dwMP3ChunkLength = 0;

	// init mad
	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);

	audio_init();

	// Install dma2-irq
    DMA2_ISR_Install();

	UpdateMP3PlayListNextChunk();

	PlayPcmNextChunk();
	GpDelay(500);

	return TRUE;
}

char* GetMP3PlayList(void)
{
	return MP3FileList[dwMP3FileCurrent].name;
}

///////////////////////////////////////////////////////////////////////////////
BOOL MP3ToPCM(const char* lpszFileName, const char* lpszPCMName)
{
	ERR_CODE result;
	F_HANDLE fOutHandle;

	int nKey;

	DWORD dwLength;
	mad_fixed_t const *ch1, *ch2;
	unsigned char pcm_data[CHUNK_LENGTH];

	// Open file
	result = GpFileOpen(lpszFileName, OPEN_R, &fHandle);
	if (result != SM_OK)
		return FALSE;

	GpFileCreate(lpszPCMName, ALWAYS_CREATE, &fOutHandle);

	pMP3ChunkBuffer = (LPBYTE) gm_malloc(MPEG_BUFSZ);
	if (!pMP3ChunkBuffer)
		return FALSE;

	//
	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);

	audio_init();

	//
	dwMP3ChunkLength = 0;

	while (1)
	{
		if (stream.next_frame)
			memmove(pMP3ChunkBuffer,
			        stream.next_frame,
			        dwMP3ChunkLength = &pMP3ChunkBuffer[dwMP3ChunkLength] - stream.next_frame);

		GpFileRead(fHandle, pMP3ChunkBuffer + dwMP3ChunkLength, MPEG_BUFSZ - dwMP3ChunkLength, &dwLength);
		mad_stream_buffer(&stream, pMP3ChunkBuffer, dwMP3ChunkLength += dwLength);

		if (!dwLength)
			break;

		while (1)
		{
			//
			if (mad_frame_decode(&frame, &stream) == -1)
				break;

			mad_synth_frame(&synth, &frame);

	  		ch1 = synth.pcm.samples[0];
	  		ch2 = synth.pcm.samples[1];

			if (synth.pcm.channels == 1)
				ch2 = ch1;

	  		dwLength = audio_pcm_s16le(pcm_data, synth.pcm.length, ch1, ch2);
	  		GpFileWrite(fOutHandle, pcm_data, dwLength);

			nKey = GpKeyGet();
			if (nKey == GPC_VK_START)
				break;
		}

		if (nKey == GPC_VK_START)
			break;
	}

 	gm_free(pMP3ChunkBuffer);

	mad_synth_finish(&synth);
	mad_frame_finish(&frame);
	mad_stream_finish(&stream);

    GpFileClose(fHandle);
    GpFileClose(fOutHandle);

    return TRUE;
}
